# -*- coding: utf-8 -*-

# Copyright 2012 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import zlib
from . import PY3

TRANSPARENT = 'TRANSPARENT'
BLACK = 'BLACK'
BLUE = 'BLUE'
RED = 'RED'
MAGENTA = 'MAGENTA'
GREEN = 'GREEN'
CYAN = 'CYAN'
YELLOW = 'YELLOW'
WHITE = 'WHITE'
BRIGHT_BLUE = 'BRIGHT_BLUE'
BRIGHT_RED = 'BRIGHT_RED'
BRIGHT_MAGENTA = 'BRIGHT_MAGENTA'
BRIGHT_GREEN = 'BRIGHT_GREEN'
BRIGHT_CYAN = 'BRIGHT_CYAN'
BRIGHT_YELLOW = 'BRIGHT_YELLOW'
BRIGHT_WHITE = 'BRIGHT_WHITE'

# Image writer options
PNG_COMPRESSION_LEVEL = 'PNGCompressionLevel'
PNG_MINIMISE_BIT_DEPTH = 'PNGMinimiseBitDepth'
PNG_USE_FAST_METHOD = 'PNGUseFastMethod'

BITS4 = [[(n & m) // m for m in (8, 4, 2, 1)] for n in range(16)]
BITS8 = [[(n & m) // m for m in (128, 64, 32, 16, 8, 4, 2, 1)] for n in range(256)]

# http://www.libpng.org/pub/png/spec/iso/index-object.html
PNG_SIGNATURE = (137, 80, 78, 71, 13, 10, 26, 10)
IHDR = (73, 72, 68, 82)
PLTE = (80, 76, 84, 69)
IDAT = (73, 68, 65, 84)
IEND_CHUNK = (0, 0, 0, 0, 73, 69, 78, 68, 174, 66, 96, 130)
CRC_MASK = 4294967295

def get_writer(palette=None, options=None, build_images=True):
    if PY3:
        return ImageWriter3(palette, options, build_images)
    return ImageWriter2(palette, options, build_images)

class ImageWriter:
    def __init__(self, palette=None, options=None, build_images=True):
        self.build_images = build_images
        if build_images:
            self.palette = self._get_default_palette()
            if palette:
                self.palette.update(palette)
            self.options = self._get_default_options()
            if options:
                for k, v in options.items():
                    try:
                        self.options[k] = int(v)
                    except ValueError:
                        pass
            self._create_colours()
            self._create_attr_dicts()
            self._create_crc_table()
            self.write_image = self._write_image
        else:
            self.write_image = self._null_write_image

    def _get_default_palette(self):
        return  {
            TRANSPARENT: (0, 254, 0),
            BLACK: (0, 0, 0),
            BLUE: (0, 0, 197),
            RED: (197, 0, 0),
            MAGENTA: (197, 0, 197),
            GREEN: (0, 198, 0),
            CYAN: (0, 198, 197),
            YELLOW: (197, 198, 0),
            WHITE: (205, 198, 205),
            BRIGHT_BLUE: (0, 0, 255),
            BRIGHT_RED: (255, 0, 0),
            BRIGHT_MAGENTA: (255, 0, 255),
            BRIGHT_GREEN: (0, 255, 0),
            BRIGHT_CYAN: (0, 255, 255),
            BRIGHT_YELLOW: (255, 255, 0),
            BRIGHT_WHITE: (255, 255, 255)
        }

    def _get_default_options(self):
        return {
            PNG_COMPRESSION_LEVEL: 9,
            PNG_MINIMISE_BIT_DEPTH: 1,
            PNG_USE_FAST_METHOD: 1
        }

    def _create_colours(self):
        colours = []
        colours.extend(self.palette[TRANSPARENT])
        colours.extend(self.palette[BLACK])
        colours.extend(self.palette[BLUE])
        colours.extend(self.palette[RED])
        colours.extend(self.palette[MAGENTA])
        colours.extend(self.palette[GREEN])
        colours.extend(self.palette[CYAN])
        colours.extend(self.palette[YELLOW])
        colours.extend(self.palette[WHITE])
        colours.extend(self.palette[BRIGHT_BLUE])
        colours.extend(self.palette[BRIGHT_RED])
        colours.extend(self.palette[BRIGHT_MAGENTA])
        colours.extend(self.palette[BRIGHT_GREEN])
        colours.extend(self.palette[BRIGHT_CYAN])
        colours.extend(self.palette[BRIGHT_YELLOW])
        colours.extend(self.palette[BRIGHT_WHITE])
        self.colours = tuple(colours)

    def _create_attr_dicts(self):
        self.attr_dict = {}
        self.attr_dict_t = {}
        self.attr_index = {}
        for attr in range(128):
            if attr & 64:
                ink = 8 + (attr & 7)
                paper = 8 + (attr & 56) // 8
                if ink == 8:
                    ink = 1
                if paper == 8:
                    paper = 1
            else:
                ink = 1 + (attr & 7)
                paper = 1 + (attr & 56) // 8
            self.attr_index[attr] = (paper, ink)
            self.attr_dict[attr] = (paper * 17, paper * 16 + ink, ink * 16 + paper, ink * 17)
            self.attr_dict_t[attr] = (0, paper, paper * 16, ink, ink * 16)

    def _create_crc_table(self):
        self.crc_table = []
        for n in range(256):
            c = n
            for k in range(8):
                if c & 1:
                    c = 3988292384 ^ (c >> 1)
                else:
                    c = c >> 1
            self.crc_table.append(c)

    def _to_bytes(self, num):
        return (num >> 24, (num >> 16) & 255, (num >> 8) & 255, num & 255)

    def _to_string(self, num):
        return ''.join(chr(b) for b in self._to_bytes(num))

    def _get_crc(self, bytes):
        crc = CRC_MASK
        for b in bytes:
            crc = self.crc_table[(crc ^ b) & 255] ^ (crc >> 8)
        return self._to_string(crc ^ CRC_MASK)

    def _get_string_crc(self, chars):
        crc = CRC_MASK
        for c in chars:
            crc = self.crc_table[(crc ^ ord(c)) & 255] ^ (crc >> 8)
        return self._to_string(crc ^ CRC_MASK)

    def _get_palette(self, udg_array, mask):
        attrs = set()
        colours = set()
        has_trans = False
        if mask:
            for row in udg_array:
                for udg in row:
                    attr = udg.attr & 127
                    attrs.add(attr)
                    paper, ink = self.attr_index[attr]
                    if udg.mask:
                        data = [udg.data[i] & udg.mask[i] for i in range(8)]
                        if any(data):
                            colours.add(ink)
                        has_paper = False
                        if has_trans:
                            for i in range(8):
                                mask_byte = udg.mask[i]
                                for j in range(8):
                                    if mask_byte & 128 == 0:
                                        colours.add(paper)
                                        has_paper = True
                                        break
                                    mask_byte *= 2
                                if has_paper:
                                    break
                        else:
                            for i in range(8):
                                byte = data[i]
                                mask_byte = udg.mask[i]
                                for j in range(8):
                                    if mask_byte & 128:
                                        if byte & 128 == 0:
                                            has_trans = True
                                    else:
                                        colours.add(paper)
                                        has_paper = True
                                    if has_paper and has_trans:
                                        break
                                    byte *= 2
                                    mask_byte *= 2
                                if has_paper and has_trans:
                                    break
                    else:
                        if any(udg.data):
                            colours.add(ink)
                        if any([byte < 255 for byte in udg.data]):
                            colours.add(paper)
                    if len(colours) > 14:
                        break
        else:
            for row in udg_array:
                for udg in row:
                    attr = udg.attr & 127
                    attrs.add(attr)
                    paper, ink = self.attr_index[attr]
                    if any(udg.data):
                        colours.add(ink)
                    if any([byte < 255 for byte in udg.data]):
                        colours.add(paper)
                    if len(colours) > 14:
                        break

        colour_map = {}
        palette = []
        i = 0
        if mask and has_trans:
            palette.extend(self.colours[0:3])
            i += 1
        for colour in colours:
            palette.extend(self.colours[colour * 3:colour * 3 + 3])
            colour_map[colour] = i
            i += 1
        palette_size = len(palette) // 3

        attr_map = {}
        for attr in attrs:
            paper, ink = self.attr_index[attr]
            attr_map[attr] = (colour_map.get(paper, 0), colour_map.get(ink, 0))

        return palette, palette_size, attr_map, has_trans

    def _build_image_data(self, udg_array, scale, mask, x, y, width, height):
        full_width = 8 * len(udg_array[0]) * scale
        full_height = 8 * len(udg_array) * scale
        width = min(width or full_width, full_width - x)
        height = min(height or full_height, full_height - y)
        full_size = width == full_width and height == full_height

        if self.options[PNG_MINIMISE_BIT_DEPTH]:
            palette, palette_size, attr_map, has_trans = self._get_palette(udg_array, mask)
            plte_chunk = self._get_plte_chunk(palette)
        else:
            if full_size:
                return width, height, 4, self.plte_chunk, self._build_image_data_bd4(udg_array, scale, mask)
            return width, height, 4, self.plte_chunk, self._build_image_data_bd_any(udg_array, scale, mask, x, y, width, height, self.attr_index, 4)

        if full_size and self.options[PNG_USE_FAST_METHOD]:
            if palette_size > 4:
                return width, height, 4, plte_chunk, self._build_image_data_bd4(udg_array, scale, has_trans, attr_map)
            if palette_size > 2:
                return width, height, 2, plte_chunk, self._build_image_data_bd2(udg_array, scale, has_trans, attr_map)
            if palette_size > 1:
                if len(udg_array) * len(udg_array[0]) > 3:
                    return width, height, 1, plte_chunk, self._build_image_data_bd1(udg_array, scale, has_trans, attr_map)
                # It's quicker to use the generic method if the UDG array is small
                return width, height, 1, plte_chunk, self._build_image_data_bd_any(udg_array, scale, has_trans, x, y, width, height, attr_map, 1)
            return width, height, 1, plte_chunk, self._build_image_data_bd1_blank(udg_array, scale)

        if palette_size > 4:
            bit_depth = 4
        elif palette_size > 2:
            bit_depth = 2
        else:
            bit_depth = 1
        return width, height, bit_depth, plte_chunk, self._build_image_data_bd_any(udg_array, scale, has_trans, x, y, width, height, attr_map, bit_depth)

    def _write_image(self, udg_array, img_file, scale=1, mask=False, x=0, y=0, width=None, height=None):
        width, height, bit_depth, plte_chunk, img_data = self._build_image_data(udg_array, scale, mask, x, y, width, height)

        # PNG signature
        img_file.write(self.png_signature)

        # IHDR
        self._write_ihdr_chunk(img_file, width, height, bit_depth)

        # PLTE
        img_file.write(plte_chunk)

        # IDAT
        self._write_idat_chunk(img_file, img_data)

        # IEND
        img_file.write(self.iend_chunk)

    def _null_write_image(self, udgs, img_file, scale=1, mask=False, x=0, y=0, width=None, height=None):
        return

class ImageWriter2(ImageWriter):
    def __init__(self, palette=None, options=None, build_images=True):
        ImageWriter.__init__(self, palette, options, build_images)
        if build_images:
            self.png_signature = ''.join(chr(b) for b in PNG_SIGNATURE)
            self.plte = ''.join(chr(b) for b in PLTE)
            self.idat = ''.join(chr(b) for b in IDAT)
            self.iend_chunk = ''.join(chr(b) for b in IEND_CHUNK)
            if not self.options[PNG_MINIMISE_BIT_DEPTH]:
                palette_str = ''.join(chr(b) for b in PLTE + self.colours)
                self.plte_chunk = self._to_string(len(palette_str) - 4)
                self.plte_chunk += palette_str
                self.plte_chunk += self._get_string_crc(palette_str)

    def _create_attr_dicts(self):
        ImageWriter._create_attr_dicts(self)
        for attr in self.attr_index:
            self.attr_dict[attr] = [chr(b) for b in self.attr_dict[attr]]
            self.attr_dict_t[attr] = [chr(b) for b in self.attr_dict_t[attr]]

    def _get_plte_chunk(self, palette):
        palette_str = self.plte + ''.join(chr(b) for b in palette)
        plte_chunk = self._to_string(len(palette))
        plte_chunk += palette_str
        plte_chunk += self._get_string_crc(palette_str)
        return plte_chunk

    def _build_image_data_bd_any(self, udg_array, scale, mask, x, y, width, height, attr_map, bit_depth):
        # Build image data at any bit depth using a generic method
        if bit_depth == 1:
            max_length = 2 + width // 8 if width & 7 else 1 + width // 8
        elif bit_depth == 2:
            max_length = 2 + width // 4 if width & 3 else 1 + width // 4
        elif bit_depth == 4:
            max_length = 2 + width // 2 if width & 1 else 1 + width // 2

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data = []
        y_count = 0
        for row in udg_array:
            if y_count >= y + height:
                break
            for i in range(8):
                if y <= y_count:
                    num_lines = min(y + height - y_count, scale)
                elif y - y_count <= scale:
                    num_lines = y - y_count
                else:
                    num_lines = 0
                y_count += scale
                if num_lines:
                    p = []
                    x_count = 0
                    for udg in row:
                        byte = udg.data[i]
                        paper, ink = attr_map[udg.attr & 127]
                        if mask and udg.mask:
                            mask_byte = udg.mask[i]
                            byte &= mask_byte
                        else:
                            mask_byte = 0
                        for b in range(8):
                            if x <= x_count:
                                num_bits = scale
                            elif x - x_count <= scale:
                                num_bits = x - x_count
                            else:
                                num_bits = 0
                            x_count += scale
                            if num_bits:
                                if byte & 128:
                                    p.extend((ink,) * num_bits)
                                elif mask_byte & 128 == 0:
                                    p.extend((paper,) * num_bits)
                                else:
                                    p.extend((0,) * num_bits)
                            byte *= 2
                            mask_byte *= 2
                    scanline = [chr(0)]
                    if bit_depth == 1:
                        p.extend((0,) * (8 - len(p) & 7))
                        scanline.extend([chr(p[j] * 128 + p[j + 1] * 64 + p[j + 2] * 32 + p[j + 3] * 16 + p[j + 4] * 8 + p[j + 5] * 4 + p[j + 6] * 2 + p[j + 7]) for j in range(0, len(p), 8)])
                    elif bit_depth == 2:
                        p.extend((0,) * (4 - len(p) & 3))
                        scanline.extend([chr(p[j] * 64 + p[j + 1] * 16 + p[j + 2] * 4 + p[j + 3]) for j in range(0, len(p), 4)])
                    elif bit_depth == 4:
                        p.extend((0,) * (2 - len(p) & 1))
                        scanline.extend([chr(p[j] * 16 + p[j + 1]) for j in range(0, len(p), 2)])
                    scanline_str = ''.join(scanline)
                    if len(scanline_str) > max_length:
                        scanline_str = scanline_str[:max_length]
                    img_data.append(compressor.compress(scanline_str * num_lines))
                if y_count >= y + height:
                    break
        img_data.append(compressor.flush())
        return ''.join(img_data)

    def _build_image_data_bd4(self, udg_array, scale, mask, attr_map=None):
        # Build image data with bit depth 4
        if attr_map:
            attr_dict = {}
            attr_dict_t = {}
            for attr, (p, i) in attr_map.items():
                attr_dict[attr] = (chr(p * 17), chr(p * 16 + i), chr(i * 16 + p), chr(i * 17))
                if mask:
                    attr_dict_t[attr] = (chr(0), chr(p), chr(p * 16), chr(i), chr(i * 16))
        else:
            attr_dict = self.attr_dict
            attr_dict_t = self.attr_dict_t

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data_blocks = []
        scale_odd = scale & 1
        h_scale = scale // 2
        for row in udg_array:
            for i in range(8):
                scanline = [chr(0)]
                for udg in row:
                    attr = udg.attr & 127
                    pp, pi, ip, ii = attr_dict[attr]
                    pph = pp * h_scale
                    iih = ii * h_scale
                    byte = udg.data[i]
                    if mask and udg.mask:
                        # Apply mask
                        tt, tp, pt, ti, it = attr_dict_t[attr]
                        tth = tt * h_scale
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        for b in range(4):
                            bits = byte & 192
                            mask_bits = mask_byte & 192
                            if bits == 192:
                                # ink + ink
                                scanline.append(ii * scale)
                            elif bits == 128:
                                scanline.append(iih)
                                if mask_bits & 64:
                                    # ink + transparent
                                    if scale_odd:
                                        scanline.append(it)
                                    scanline.append(tth)
                                else:
                                    # ink + paper
                                    if scale_odd:
                                        scanline.append(ip)
                                    scanline.append(pph)
                            elif bits == 64:
                                if mask_bits & 128:
                                    # transparent + ink
                                    scanline.append(tth)
                                    if scale_odd:
                                        scanline.append(ti)
                                else:
                                    # paper + ink
                                    scanline.append(pph)
                                    if scale_odd:
                                        scanline.append(pi)
                                scanline.append(iih)
                            else:
                                if mask_bits == 192:
                                    # transparent + transparent
                                    scanline.append(tt * scale)
                                elif mask_bits == 128:
                                    # transparent + paper
                                    scanline.append(tth)
                                    if scale_odd:
                                        scanline.append(tp)
                                    scanline.append(pph)
                                elif mask_bits == 64:
                                    # paper + transparent
                                    scanline.append(pph)
                                    if scale_odd:
                                        scanline.append(pt)
                                    scanline.append(tth)
                                else:
                                    # paper + paper
                                    scanline.append(pp * scale)
                            byte *= 4
                            mask_byte *= 4
                    else:
                        # No mask
                        if byte == 0:
                            scanline.append(pp * 4 * scale)
                        elif byte == 255:
                            scanline.append(ii * 4 * scale)
                        else:
                            for b in range(4):
                                bits = byte & 192
                                if bits == 0:
                                    # paper + paper
                                    scanline.append(pp * scale)
                                elif bits == 192:
                                    # ink + ink
                                    scanline.append(ii * scale)
                                elif bits == 128:
                                    # ink + paper
                                    scanline.append(iih)
                                    if scale_odd:
                                        scanline.append(ip)
                                    scanline.append(pph)
                                else:
                                    # paper + ink
                                    scanline.append(pph)
                                    if scale_odd:
                                        scanline.append(pi)
                                    scanline.append(iih)
                                byte *= 4
                img_data_blocks.append(compressor.compress(''.join(scanline) * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd2(self, udg_array, scale, mask, attr_map):
        # Build image data with bit depth 2
        attr_dict = {}
        for attr, q in attr_map.items():
            attr_dict[attr] = [chr(q[b3] * 64 + q[b2] * 16 + q[b1] * 4 + q[b0]) for b3, b2, b1, b0 in BITS4]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        scale_m = scale & 3
        q_scale = scale // 4
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = [chr(0)]
                for udg in row:
                    attr = udg.attr & 127
                    byte = udg.data[i]
                    attrs = attr_dict[attr]
                    if mask and udg.mask:
                        # Apply mask
                        p = []
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        paper, ink = attr_map[attr]
                        for b in range(8):
                            if byte & 128:
                                p.extend((ink,) * scale)
                            elif mask_byte & 128:
                                p.extend((0,) * scale)
                            else:
                                p.extend((paper,) * scale)
                            byte *= 2
                            mask_byte *= 2
                        scanline.extend([chr(p[j] * 64 + p[j + 1] * 16 + p[j + 2] * 4 + p[j + 3]) for j in range(0, len(p), 4)])
                    else:
                        # No mask
                        for bits in ((byte & 240) // 16, byte & 15):
                            b3, b2, b1, b0 = BITS4[bits]
                            if scale_m == 0:
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b2 * 15] * q_scale)
                                scanline.append(attrs[b1 * 15] * q_scale)
                                scanline.append(attrs[b0 * 15] * q_scale)
                            elif scale_m == 1:
                                if scale == 1:
                                    scanline.append(attrs[bits])
                                    continue
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b3 * 8 + b2 * 7])
                                scanline.append(attrs[b2 * 15] * (q_scale - 1))
                                scanline.append(attrs[b2 * 12 + b1 * 3])
                                scanline.append(attrs[b1 * 15] * (q_scale - 1))
                                scanline.append(attrs[b1 * 14 + b0])
                                scanline.append(attrs[b0 * 15] * q_scale)
                            elif scale_m == 2:
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b3 * 12 + b2 * 3])
                                scanline.append(attrs[b2 * 15] * q_scale)
                                scanline.append(attrs[b1 * 15] * q_scale)
                                scanline.append(attrs[b1 * 12 + b0 * 3])
                                scanline.append(attrs[b0 * 15] * q_scale)
                            else:
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b3 * 14 + b2])
                                scanline.append(attrs[b2 * 15] * q_scale)
                                scanline.append(attrs[b2 * 12 + b1 * 3])
                                scanline.append(attrs[b1 * 15] * q_scale)
                                scanline.append(attrs[b1 * 8 + b0 * 7])
                                scanline.append(attrs[b0 * 15] * q_scale)
                img_data_blocks.append(compressor.compress(''.join(scanline) * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd1(self, udg_array, scale, mask, attr_map):
        # Build image data with bit depth 1 and a 2-colour palette
        attr_dict = {}
        for attr, q in attr_map.items():
            attr_dict[attr] = [chr(q[b7] * 128 + q[b6] * 64 + q[b5] * 32 + q[b4] * 16 + q[b3] * 8 + q[b2] * 4 + q[b1] * 2 + q[b0]) for b7, b6, b5, b4, b3, b2, b1, b0 in BITS8]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        e_scale = scale // 8
        scale_m = scale & 7
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = [chr(0)]
                for udg in row:
                    attr = udg.attr & 127
                    byte = udg.data[i]
                    attrs = attr_dict[attr]
                    if mask and udg.mask:
                        # Apply mask
                        p = []
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        paper, ink = attr_map[attr]
                        for b in range(8):
                            if byte & 128:
                                p.extend((ink,) * scale)
                            elif mask_byte & 128:
                                p.extend((0,) * scale)
                            else:
                                p.extend((paper,) * scale)
                            byte *= 2
                            mask_byte *= 2
                        scanline.extend([chr(p[j] * 128 + p[j + 1] * 64 + p[j + 2] * 32 + p[j + 3] * 16 + p[j + 4] * 8 + p[j + 5] * 4 + p[j + 6] * 2 + p[j + 7]) for j in range(0, len(p), 8)])
                    else:
                        # No mask
                        b7, b6, b5, b4, b3, b2, b1, b0 = BITS8[byte]
                        attrs = attr_dict[udg.attr & 127]
                        if scale_m == 0:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 1:
                            if scale == 1:
                                scanline.append(attrs[byte])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 128 + b6 * 127])
                            scanline.append(attrs[b6 * 255] * (e_scale - 1))
                            scanline.append(attrs[b6 * 192 + b5 * 63])
                            scanline.append(attrs[b5 * 255] * (e_scale - 1))
                            scanline.append(attrs[b5 * 224 + b4 * 31])
                            scanline.append(attrs[b4 * 255] * (e_scale - 1))
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * (e_scale - 1))
                            scanline.append(attrs[b3 * 248 + b2 * 7])
                            scanline.append(attrs[b2 * 255] * (e_scale - 1))
                            scanline.append(attrs[b2 * 252 + b1 * 3])
                            scanline.append(attrs[b1 * 255] * (e_scale - 1))
                            scanline.append(attrs[b1 * 254 + b0])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 2:
                            if scale == 2:
                                scanline.append(attrs[b7 * 192 + b6 * 48 + b5 * 12 + b4 * 3])
                                scanline.append(attrs[b3 * 192 + b2 * 48 + b1 * 12 + b0 * 3])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 192 + b6 * 63])
                            scanline.append(attrs[b6 * 255] * (e_scale - 1))
                            scanline.append(attrs[b6 * 240 + b5 * 15])
                            scanline.append(attrs[b5 * 255] * (e_scale - 1))
                            scanline.append(attrs[b5 * 252 + b4 * 3])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 192 + b2 * 63])
                            scanline.append(attrs[b2 * 255] * (e_scale - 1))
                            scanline.append(attrs[b2 * 240 + b1 * 15])
                            scanline.append(attrs[b1 * 255] * (e_scale - 1))
                            scanline.append(attrs[b1 * 252 + b0 * 3])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 3:
                            if scale == 3:
                                scanline.append(attrs[b7 * 224 + b6 * 28 + b5 * 3])
                                scanline.append(attrs[b5 * 128 + b4 * 112 + b3 * 14 + b2])
                                scanline.append(attrs[b2 * 192 + b1 * 56 + b0 * 7])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 224 + b6 * 31])
                            scanline.append(attrs[b6 * 255] * (e_scale - 1))
                            scanline.append(attrs[b6 * 252 + b5 * 3])
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 128 + b4 * 127])
                            scanline.append(attrs[b4 * 255] * (e_scale - 1))
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * (e_scale - 1))
                            scanline.append(attrs[b3 * 254 + b2])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b2 * 192 + b1 * 63])
                            scanline.append(attrs[b1 * 255] * (e_scale - 1))
                            scanline.append(attrs[b1 * 248 + b0 * 7])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 4:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 240 + b6 * 15])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 240 + b4 * 15])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 240 + b2 * 15])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 240 + b0 * 15])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 5:
                            if scale == 5:
                                scanline.append(attrs[b7 * 248 + b6 * 7])
                                scanline.append(attrs[b6 * 192 + b5 * 62 + b4])
                                scanline.append(attrs[b4 * 240 + b3 * 15])
                                scanline.append(attrs[b3 * 128 + b2 * 124 + b1 * 3])
                                scanline.append(attrs[b1 * 224 + b0 * 31])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 248 + b6 * 7])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b6 * 192 + b5 * 63])
                            scanline.append(attrs[b5 * 255] * (e_scale - 1))
                            scanline.append(attrs[b5 * 254 + b4])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 128 + b2 * 127])
                            scanline.append(attrs[b2 * 255] * (e_scale - 1))
                            scanline.append(attrs[b2 * 252 + b1 * 3])
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 224 + b0 * 31])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 6:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 252 + b6 * 3])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b6 * 240 + b5 * 15])
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 192 + b4 * 63])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 252 + b2 * 3])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b2 * 240 + b1 * 15])
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 192 + b0 * 63])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 7:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 254 + b6])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b6 * 252 + b5 * 3])
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 248 + b4 * 7])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 224 + b2 * 31])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b2 * 192 + b1 * 63])
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 128 + b0 * 127])
                            scanline.append(attrs[b0 * 255] * e_scale)
                img_data_blocks.append(compressor.compress(''.join(scanline) * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd1_blank(self, udg_array, scale):
        # Build image data with bit depth 1 and a single-colour palette;
        # placing the integer value in brackets means it is evaluated before
        # 'multiplying' the string (and is therefore quicker)
        scanlines = chr(0) * ((1 + len(udg_array[0]) * scale) * len(udg_array) * scale * 8)
        return zlib.compress(scanlines, self.options[PNG_COMPRESSION_LEVEL])

    def _write_ihdr_chunk(self, img_file, width, height, bit_depth):
        data = list(IHDR) # chunk type
        data.extend(self._to_bytes(width)) # width
        data.extend(self._to_bytes(height)) # height
        data.append(bit_depth) # bit depth
        data.append(3) # colour type
        data.extend((0, 0, 0)) # compression, filter and interlace methods
        img_file.write(self._to_string(len(data) - 4)) # chunk length
        img_file.write(''.join(chr(b) for b in data))
        img_file.write(self._get_crc(data)) # CRC

    def _write_idat_chunk(self, img_file, img_data):
        img_file.write(self._to_string(len(img_data))) # chunk length
        img_data = self.idat + img_data
        img_file.write(img_data) # chunk type and data
        img_file.write(self._get_string_crc(img_data)) # CRC

class ImageWriter3(ImageWriter):
    def __init__(self, palette=None, options=None, build_images=True):
        ImageWriter.__init__(self, palette, options, build_images)
        if build_images:
            self.png_signature = bytes(PNG_SIGNATURE)
            self.idat = bytes(IDAT)
            self.iend_chunk = bytes(IEND_CHUNK)
            if not self.options[PNG_MINIMISE_BIT_DEPTH]:
                palette = PLTE + self.colours
                self.plte_chunk = bytearray(self._to_bytes(len(palette) - 4))
                self.plte_chunk.extend(palette)
                self.plte_chunk.extend(self._get_crc(palette))

    def _get_crc(self, byte_list):
        crc = CRC_MASK
        for b in byte_list:
            crc = self.crc_table[(crc ^ b) & 255] ^ (crc >> 8)
        return self._to_bytes(crc ^ CRC_MASK)

    def _get_plte_chunk(self, palette):
        palette_bytes = []
        palette_bytes.extend(PLTE)
        palette_bytes.extend(palette)
        plte_chunk = bytearray(self._to_bytes(len(palette)))
        plte_chunk.extend(palette_bytes)
        plte_chunk.extend(self._get_crc(palette_bytes))
        return plte_chunk

    def _build_image_data_bd_any(self, udg_array, scale, mask, x, y, width, height, attr_map, bit_depth):
        # Build image data at any bit depth using a generic method
        if bit_depth == 1:
            max_length = 2 + width // 8 if width & 7 else 1 + width // 8
        elif bit_depth == 2:
            max_length = 2 + width // 4 if width & 3 else 1 + width // 4
        elif bit_depth == 4:
            max_length = 2 + width // 2 if width & 1 else 1 + width // 2

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data = bytearray()
        y_count = 0
        for row in udg_array:
            if y_count >= y + height:
                break
            for i in range(8):
                if y <= y_count:
                    num_lines = min(y + height - y_count, scale)
                elif y - y_count <= scale:
                    num_lines = y - y_count
                else:
                    num_lines = 0
                y_count += scale
                if num_lines:
                    p = []
                    x_count = 0
                    for udg in row:
                        byte = udg.data[i]
                        paper, ink = attr_map[udg.attr & 127]
                        if mask and udg.mask:
                            mask_byte = udg.mask[i]
                            byte &= mask_byte
                        else:
                            mask_byte = 0
                        for b in range(8):
                            if x <= x_count:
                                num_bits = scale
                            elif x - x_count <= scale:
                                num_bits = x - x_count
                            else:
                                num_bits = 0
                            x_count += scale
                            if num_bits:
                                if byte & 128:
                                    p.extend((ink,) * num_bits)
                                elif mask_byte & 128 == 0:
                                    p.extend((paper,) * num_bits)
                                else:
                                    p.extend((0,) * num_bits)
                            byte *= 2
                            mask_byte *= 2
                    scanline = bytearray((0,))
                    if bit_depth == 1:
                        p.extend((0,) * (8 - len(p) & 7))
                        scanline.extend([p[j] * 128 + p[j + 1] * 64 + p[j + 2] * 32 + p[j + 3] * 16 + p[j + 4] * 8 + p[j + 5] * 4 + p[j + 6] * 2 + p[j + 7] for j in range(0, len(p), 8)])
                    elif bit_depth == 2:
                        p.extend((0,) * (4 - len(p) & 3))
                        scanline.extend([p[j] * 64 + p[j + 1] * 16 + p[j + 2] * 4 + p[j + 3] for j in range(0, len(p), 4)])
                    elif bit_depth == 4:
                        p.extend((0,) * (2 - len(p) & 1))
                        scanline.extend([p[j] * 16 + p[j + 1] for j in range(0, len(p), 2)])
                    if len(scanline) > max_length:
                        scanline = scanline[:max_length]
                    img_data.extend(compressor.compress(scanline * num_lines))
                if y_count >= y + height:
                    break
        img_data.extend(compressor.flush())
        return img_data

    def _build_image_data_bd4(self, udg_array, scale, mask, attr_map=None):
        # Build image data with bit depth 4
        if attr_map:
            attr_dict = {}
            attr_dict_t = {}
            for attr, (p, i) in attr_map.items():
                attr_dict[attr] = (p * 17, p * 16 + i, i * 16 + p, i * 17)
                if mask:
                    attr_dict_t[attr] = (0, p, p * 16, i, i * 16)
        else:
            attr_dict = self.attr_dict
            attr_dict_t = self.attr_dict_t

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data = bytearray()
        scale_odd = scale & 1
        h_scale = scale // 2
        for row in udg_array:
            for i in range(8):
                scanline = bytearray((0,))
                for udg in row:
                    attr = udg.attr & 127
                    pp, pi, ip, ii = attr_dict[attr]
                    pph = (pp,) * h_scale
                    iih = (ii,) * h_scale
                    byte = udg.data[i]
                    if mask and udg.mask:
                        # Apply mask
                        tt, tp, pt, ti, it = attr_dict_t[attr]
                        tth = (tt,) * h_scale
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        for b in range(4):
                            bits = byte & 192
                            mask_bits = mask_byte & 192
                            if bits == 192:
                                # ink + ink
                                scanline.extend((ii,) * scale)
                            elif bits == 128:
                                scanline.extend(iih)
                                if mask_bits & 64:
                                    # ink + transparent
                                    if scale_odd:
                                        scanline.append(it)
                                    scanline.extend(tth)
                                else:
                                    # ink + paper
                                    if scale_odd:
                                        scanline.append(ip)
                                    scanline.extend(pph)
                            elif bits == 64:
                                if mask_bits & 128:
                                    # transparent + ink
                                    scanline.extend(tth)
                                    if scale_odd:
                                        scanline.append(ti)
                                else:
                                    # paper + ink
                                    scanline.extend(pph)
                                    if scale_odd:
                                        scanline.append(pi)
                                scanline.extend(iih)
                            else:
                                if mask_bits == 192:
                                    # transparent + transparent
                                    scanline.extend((tt,) * scale)
                                elif mask_bits == 128:
                                    # transparent + paper
                                    scanline.extend(tth)
                                    if scale_odd:
                                        scanline.append(tp)
                                    scanline.extend(pph)
                                elif mask_bits == 64:
                                    # paper + transparent
                                    scanline.extend(pph)
                                    if scale_odd:
                                        scanline.append(pt)
                                    scanline.extend(tth)
                                else:
                                    # paper + paper
                                    scanline.extend((pp,) * scale)
                            byte *= 4
                            mask_byte *= 4
                    else:
                        # No mask
                        if byte == 0:
                            scanline.extend((pp,) * 4 * scale)
                        elif byte == 255:
                            scanline.extend((ii,) * 4 * scale)
                        else:
                            for b in range(4):
                                bits = byte & 192
                                if bits == 0:
                                    # paper + paper
                                    scanline.extend((pp,) * scale)
                                elif bits == 192:
                                    # ink + ink
                                    scanline.extend((ii,) * scale)
                                elif bits == 128:
                                    # ink + paper
                                    scanline.extend(iih)
                                    if scale_odd:
                                        scanline.append(ip)
                                    scanline.extend(pph)
                                else:
                                    # paper + ink
                                    scanline.extend(pph)
                                    if scale_odd:
                                        scanline.append(pi)
                                    scanline.extend(iih)
                                byte *= 4
                img_data.extend(compressor.compress(scanline * scale))
        img_data.extend(compressor.flush())
        return img_data

    def _build_image_data_bd2(self, udg_array, scale, mask, attr_map):
        # Build image data with bit depth 2
        attr_dict = {}
        for attr, q in attr_map.items():
            attr_dict[attr] = [q[b3] * 64 + q[b2] * 16 + q[b1] * 4 + q[b0] for b3, b2, b1, b0 in BITS4]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        scale_m = scale & 3
        q_scale = scale // 4
        img_data = bytearray()
        for row in udg_array:
            for i in range(8):
                scanline = bytearray((0,))
                for udg in row:
                    attr = udg.attr & 127
                    byte = udg.data[i]
                    attrs = attr_dict[attr]
                    if mask and udg.mask:
                        # Apply mask
                        p = []
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        paper, ink = attr_map[attr]
                        for b in range(8):
                            if byte & 128:
                                p.extend((ink,) * scale)
                            elif mask_byte & 128:
                                p.extend((0,) * scale)
                            else:
                                p.extend((paper,) * scale)
                            byte *= 2
                            mask_byte *= 2
                        scanline.extend([p[j] * 64 + p[j + 1] * 16 + p[j + 2] * 4 + p[j + 3] for j in range(0, len(p), 4)])
                    else:
                        # No mask
                        for bits in ((byte & 240) // 16, byte & 15):
                            b3, b2, b1, b0 = BITS4[bits]
                            if scale_m == 0:
                                scanline.extend((attrs[b3 * 15],) * q_scale)
                                scanline.extend((attrs[b2 * 15],) * q_scale)
                                scanline.extend((attrs[b1 * 15],) * q_scale)
                                scanline.extend((attrs[b0 * 15],) * q_scale)
                            elif scale_m == 1:
                                if scale == 1:
                                    scanline.append(attrs[bits])
                                    continue
                                scanline.extend((attrs[b3 * 15],) * q_scale)
                                scanline.append(attrs[b3 * 8 + b2 * 7])
                                scanline.extend((attrs[b2 * 15],) * (q_scale - 1))
                                scanline.append(attrs[b2 * 12 + b1 * 3])
                                scanline.extend((attrs[b1 * 15],) * (q_scale - 1))
                                scanline.append(attrs[b1 * 14 + b0])
                                scanline.extend((attrs[b0 * 15],) * q_scale)
                            elif scale_m == 2:
                                scanline.extend((attrs[b3 * 15],) * q_scale)
                                scanline.append(attrs[b3 * 12 + b2 * 3])
                                scanline.extend((attrs[b2 * 15],) * q_scale)
                                scanline.extend((attrs[b1 * 15],) * q_scale)
                                scanline.append(attrs[b1 * 12 + b0 * 3])
                                scanline.extend((attrs[b0 * 15],) * q_scale)
                            else:
                                scanline.extend((attrs[b3 * 15],) * q_scale)
                                scanline.append(attrs[b3 * 14 + b2])
                                scanline.extend((attrs[b2 * 15],) * q_scale)
                                scanline.append(attrs[b2 * 12 + b1 * 3])
                                scanline.extend((attrs[b1 * 15],) * q_scale)
                                scanline.append(attrs[b1 * 8 + b0 * 7])
                                scanline.extend((attrs[b0 * 15],) * q_scale)
                img_data.extend(compressor.compress(scanline * scale))
        img_data.extend(compressor.flush())
        return img_data

    def _build_image_data_bd1(self, udg_array, scale, mask, attr_map):
        # Build image data with bit depth 1 and a 2-colour palette
        attr_dict = {}
        for attr, q in attr_map.items():
            attr_dict[attr] = [q[b7] * 128 + q[b6] * 64 + q[b5] * 32 + q[b4] * 16 + q[b3] * 8 + q[b2] * 4 + q[b1] * 2 + q[b0] for b7, b6, b5, b4, b3, b2, b1, b0 in BITS8]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        e_scale = scale // 8
        scale_m = scale & 7
        img_data = bytearray()
        for row in udg_array:
            for i in range(8):
                scanline = bytearray((0,))
                for udg in row:
                    attr = udg.attr & 127
                    byte = udg.data[i]
                    attrs = attr_dict[attr]
                    if mask and udg.mask:
                        # Apply mask
                        p = []
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        paper, ink = attr_map[attr]
                        for b in range(8):
                            if byte & 128:
                                p.extend((ink,) * scale)
                            elif mask_byte & 128:
                                p.extend((0,) * scale)
                            else:
                                p.extend((paper,) * scale)
                            byte *= 2
                            mask_byte *= 2
                        scanline.extend([p[j] * 128 + p[j + 1] * 64 + p[j + 2] * 32 + p[j + 3] * 16 + p[j + 4] * 8 + p[j + 5] * 4 + p[j + 6] * 2 + p[j + 7] for j in range(0, len(p), 8)])
                    else:
                        # No mask
                        b7, b6, b5, b4, b3, b2, b1, b0 = BITS8[byte]
                        attrs = attr_dict[udg.attr & 127]
                        if scale_m == 0:
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.extend((attrs[b6 * 255],) * e_scale)
                            scanline.extend((attrs[b5 * 255],) * e_scale)
                            scanline.extend((attrs[b4 * 255],) * e_scale)
                            scanline.extend((attrs[b3 * 255],) * e_scale)
                            scanline.extend((attrs[b2 * 255],) * e_scale)
                            scanline.extend((attrs[b1 * 255],) * e_scale)
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 1:
                            if scale == 1:
                                scanline.append(attrs[byte])
                                continue
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 128 + b6 * 127])
                            scanline.extend((attrs[b6 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b6 * 192 + b5 * 63])
                            scanline.extend((attrs[b5 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b5 * 224 + b4 * 31])
                            scanline.extend((attrs[b4 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.extend((attrs[b3 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b3 * 248 + b2 * 7])
                            scanline.extend((attrs[b2 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b2 * 252 + b1 * 3])
                            scanline.extend((attrs[b1 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b1 * 254 + b0])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 2:
                            if scale == 2:
                                scanline.append(attrs[b7 * 192 + b6 * 48 + b5 * 12 + b4 * 3])
                                scanline.append(attrs[b3 * 192 + b2 * 48 + b1 * 12 + b0 * 3])
                                continue
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 192 + b6 * 63])
                            scanline.extend((attrs[b6 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b6 * 240 + b5 * 15])
                            scanline.extend((attrs[b5 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b5 * 252 + b4 * 3])
                            scanline.extend((attrs[b4 * 255],) * e_scale)
                            scanline.extend((attrs[b3 * 255],) * e_scale)
                            scanline.append(attrs[b3 * 192 + b2 * 63])
                            scanline.extend((attrs[b2 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b2 * 240 + b1 * 15])
                            scanline.extend((attrs[b1 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b1 * 252 + b0 * 3])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 3:
                            if scale == 3:
                                scanline.append(attrs[b7 * 224 + b6 * 28 + b5 * 3])
                                scanline.append(attrs[b5 * 128 + b4 * 112 + b3 * 14 + b2])
                                scanline.append(attrs[b2 * 192 + b1 * 56 + b0 * 7])
                                continue
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 224 + b6 * 31])
                            scanline.extend((attrs[b6 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b6 * 252 + b5 * 3])
                            scanline.extend((attrs[b5 * 255],) * e_scale)
                            scanline.append(attrs[b5 * 128 + b4 * 127])
                            scanline.extend((attrs[b4 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.extend((attrs[b3 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b3 * 254 + b2])
                            scanline.extend((attrs[b2 * 255],) * e_scale)
                            scanline.append(attrs[b2 * 192 + b1 * 63])
                            scanline.extend((attrs[b1 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b1 * 248 + b0 * 7])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 4:
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 240 + b6 * 15])
                            scanline.extend((attrs[b6 * 255],) * e_scale)
                            scanline.extend((attrs[b5 * 255],) * e_scale)
                            scanline.append(attrs[b5 * 240 + b4 * 15])
                            scanline.extend((attrs[b4 * 255],) * e_scale)
                            scanline.extend((attrs[b3 * 255],) * e_scale)
                            scanline.append(attrs[b3 * 240 + b2 * 15])
                            scanline.extend((attrs[b2 * 255],) * e_scale)
                            scanline.extend((attrs[b1 * 255],) * e_scale)
                            scanline.append(attrs[b1 * 240 + b0 * 15])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 5:
                            if scale == 5:
                                scanline.append(attrs[b7 * 248 + b6 * 7])
                                scanline.append(attrs[b6 * 192 + b5 * 62 + b4])
                                scanline.append(attrs[b4 * 240 + b3 * 15])
                                scanline.append(attrs[b3 * 128 + b2 * 124 + b1 * 3])
                                scanline.append(attrs[b1 * 224 + b0 * 31])
                                continue
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 248 + b6 * 7])
                            scanline.extend((attrs[b6 * 255],) * e_scale)
                            scanline.append(attrs[b6 * 192 + b5 * 63])
                            scanline.extend((attrs[b5 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b5 * 254 + b4])
                            scanline.extend((attrs[b4 * 255],) * e_scale)
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.extend((attrs[b3 * 255],) * e_scale)
                            scanline.append(attrs[b3 * 128 + b2 * 127])
                            scanline.extend((attrs[b2 * 255],) * (e_scale - 1))
                            scanline.append(attrs[b2 * 252 + b1 * 3])
                            scanline.extend((attrs[b1 * 255],) * e_scale)
                            scanline.append(attrs[b1 * 224 + b0 * 31])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 6:
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 252 + b6 * 3])
                            scanline.extend((attrs[b6 * 255],) * e_scale)
                            scanline.append(attrs[b6 * 240 + b5 * 15])
                            scanline.extend((attrs[b5 * 255],) * e_scale)
                            scanline.append(attrs[b5 * 192 + b4 * 63])
                            scanline.extend((attrs[b4 * 255],) * e_scale)
                            scanline.extend((attrs[b3 * 255],) * e_scale)
                            scanline.append(attrs[b3 * 252 + b2 * 3])
                            scanline.extend((attrs[b2 * 255],) * e_scale)
                            scanline.append(attrs[b2 * 240 + b1 * 15])
                            scanline.extend((attrs[b1 * 255],) * e_scale)
                            scanline.append(attrs[b1 * 192 + b0 * 63])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                        elif scale_m == 7:
                            scanline.extend((attrs[b7 * 255],) * e_scale)
                            scanline.append(attrs[b7 * 254 + b6])
                            scanline.extend((attrs[b6 * 255],) * e_scale)
                            scanline.append(attrs[b6 * 252 + b5 * 3])
                            scanline.extend((attrs[b5 * 255],) * e_scale)
                            scanline.append(attrs[b5 * 248 + b4 * 7])
                            scanline.extend((attrs[b4 * 255],) * e_scale)
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.extend((attrs[b3 * 255],) * e_scale)
                            scanline.append(attrs[b3 * 224 + b2 * 31])
                            scanline.extend((attrs[b2 * 255],) * e_scale)
                            scanline.append(attrs[b2 * 192 + b1 * 63])
                            scanline.extend((attrs[b1 * 255],) * e_scale)
                            scanline.append(attrs[b1 * 128 + b0 * 127])
                            scanline.extend((attrs[b0 * 255],) * e_scale)
                img_data.extend(compressor.compress(scanline * scale))
        img_data.extend(compressor.flush())
        return img_data

    def _build_image_data_bd1_blank(self, udg_array, scale):
        # Build image data with bit depth 1 and a single-colour palette;
        # placing the integer value in brackets means it is evaluated before
        # 'multiplying' the tuple (and is therefore quicker)
        scanlines = bytearray((0,) * ((1 + len(udg_array[0]) * scale) * len(udg_array) * scale * 8))
        return zlib.compress(scanlines, self.options[PNG_COMPRESSION_LEVEL])

    def _write_ihdr_chunk(self, img_file, width, height, bit_depth):
        data = bytearray(IHDR) # chunk type
        data.extend(bytes(self._to_bytes(width))) # width
        data.extend(bytes(self._to_bytes(height))) # height
        data.extend(bytes((bit_depth, 3))) # bit depth and colour type
        data.extend(bytes((0, 0, 0))) # compression, filter and interlace methods
        img_file.write(bytes(self._to_bytes(len(data) - 4))) # chunk length
        img_file.write(data)
        img_file.write(bytes(self._get_crc(data))) # CRC

    def _write_idat_chunk(self, img_file, img_data):
        img_file.write(bytes(self._to_bytes(len(img_data)))) # chunk length
        img_data = self.idat + img_data
        img_file.write(img_data) # chunk type and data
        img_file.write(bytes(self._get_crc(img_data))) # CRC
