# -*- coding: utf-8 -*-
import zlib
import unittest
from skoolkittest import SkoolKitTestCase, PY2
if PY2:
    from StringIO import StringIO as StreamIO
else:
    from io import BytesIO as StreamIO
from skoolkit.image import get_writer, PNG_COMPRESSION_LEVEL, PNG_MINIMISE_BIT_DEPTH, PNG_USE_FAST_METHOD
from skoolkit.skoolhtml import Udg

TRANSPARENT = [0, 254, 0]
BLACK = [0, 0, 0]
BLUE = [0, 0, 197]
RED = [197, 0, 0]
MAGENTA = [197, 0, 197]
GREEN = [0, 198, 0]
CYAN = [0, 198, 197]
YELLOW = [197, 198, 0]
WHITE = [205, 198, 205]
BRIGHT_BLUE = [0, 0, 255]
BRIGHT_RED = [255, 0, 0]
BRIGHT_MAGENTA = [255, 0, 255]
BRIGHT_GREEN = [0, 255, 0]
BRIGHT_CYAN = [0, 255, 255]
BRIGHT_YELLOW = [255, 255, 0]
BRIGHT_WHITE = [255, 255, 255]

PALETTE = [
    TRANSPARENT,
    BLACK,
    BLUE,
    RED,
    MAGENTA,
    GREEN,
    CYAN,
    YELLOW,
    WHITE,
    BRIGHT_BLUE,
    BRIGHT_RED,
    BRIGHT_MAGENTA,
    BRIGHT_GREEN,
    BRIGHT_CYAN,
    BRIGHT_YELLOW,
    BRIGHT_WHITE
]

PNG_SIGNATURE = [137, 80, 78, 71, 13, 10, 26, 10]
IHDR = [73, 72, 68, 82]
PLTE = [80, 76, 84, 69]
IDAT = [73, 68, 65, 84]
IEND = [0, 0, 0, 0, 73, 69, 78, 68, 174, 66, 96, 130]

CRC_MASK = 4294967295

def create_crc_table():
    crc_table = []
    for n in range(256):
        c = n
        for k in range(8):
            if c & 1:
                c = 3988292384 ^ (c >> 1)
            else:
                c = c >> 1
        crc_table.append(c)
    return crc_table

def create_attr_index():
    attr_index = {}
    for attr in range(128):
        if attr & 64:
            ink = 8 + (attr & 7)
            paper = 8 + (attr & 56) // 8
            if ink == 8:
                ink = 1
            if paper == 8:
                paper = 1
        else:
            ink = 1 + (attr & 7)
            paper = 1 + (attr & 56) // 8
        attr_index[attr] = (paper, ink)
    return attr_index

CRC_TABLE = create_crc_table()
ATTR_INDEX = create_attr_index()

class ImageWriterTest(SkoolKitTestCase):
    def _get_crc(self, byte_list):
        crc = CRC_MASK
        for b in byte_list:
            crc = CRC_TABLE[(crc ^ b) & 255] ^ (crc >> 8)
        return crc ^ CRC_MASK

    def _get_num(self, stream, index):
        return index + 1, stream[index]

    def _get_dword(self, stream, index):
        return index + 4, 16777216 * stream[index] + 65536 * stream[index + 1] + 256 * stream[index + 2] + stream[index + 3]

    def _get_chunk_type(self, stream, index):
        return index + 4, stream[index:index + 4]

    def _check_signature(self, img_bytes):
        i = len(PNG_SIGNATURE)
        self.assertEqual(img_bytes[:i], PNG_SIGNATURE)
        return i

    def _check_ihdr(self, img_bytes, index, exp_width, exp_height, exp_bit_depth):
        i = index
        i, chunk_length = self._get_dword(img_bytes, i)
        self.assertEqual(chunk_length, 13)
        ihdr_start = i
        i, chunk_type = self._get_chunk_type(img_bytes, i)
        self.assertEqual(chunk_type, IHDR)
        i, image_width = self._get_dword(img_bytes, i)
        self.assertEqual(image_width, exp_width)
        i, image_height = self._get_dword(img_bytes, i)
        self.assertEqual(image_height, exp_height)
        i, bit_depth = self._get_num(img_bytes, i)
        self.assertEqual(bit_depth, exp_bit_depth)
        i, colour_type = self._get_num(img_bytes, i)
        self.assertEqual(colour_type, 3)
        i, compression_method = self._get_num(img_bytes, i)
        self.assertEqual(compression_method, 0)
        i, filter_method = self._get_num(img_bytes, i)
        self.assertEqual(filter_method, 0)
        i, interlace_method = self._get_num(img_bytes, i)
        self.assertEqual(interlace_method, 0)
        ihdr_end = i
        i, ihdr_crc = self._get_dword(img_bytes, i)
        self.assertEqual(ihdr_crc, self._get_crc(img_bytes[ihdr_start:ihdr_end]))
        return i, image_width, bit_depth

    def _check_plte(self, img_bytes, index, exp_palette):
        i = index
        i, chunk_length = self._get_dword(img_bytes, i)
        self.assertEqual(chunk_length, len(exp_palette) * 3)
        plte_start = i
        i, chunk_type = self._get_chunk_type(img_bytes, i)
        self.assertEqual(chunk_type, PLTE)
        palette = [img_bytes[j:j + 3] for j in range(i, i + chunk_length, 3)]
        i += chunk_length
        self.assertEqual(sorted(palette), sorted(exp_palette))
        plte_end = i
        i, plte_crc = self._get_dword(img_bytes, i)
        self.assertEqual(plte_crc, self._get_crc(img_bytes[plte_start:plte_end]))
        return i, palette

    def _get_pixels_from_image_data(self, bit_depth, palette, image_data, width):
        if bit_depth == 4:
            scanline_len = 2 + width // 2 if width & 1 else 1 + width // 2
        elif bit_depth == 2:
            scanline_len = 2 + width // 4 if width & 3 else 1 + width // 4
        else:
            scanline_len = 2 + width // 8 if width & 7 else 1 + width // 8
        pixels = []
        for i, byte in enumerate(image_data):
            if i % scanline_len == 0:
                self.assertEqual(byte, 0)
                if i:
                    pixels.append(pixel_row[:width])
                pixel_row = []
            elif bit_depth == 4:
                pixel_row.append(palette[(byte & 240) // 16])
                pixel_row.append(palette[byte & 15])
            elif bit_depth == 2:
                for b in range(4):
                    pixel_row.append(palette[(byte & 192) // 64])
                    byte *= 4
            else:
                for b in range(8):
                    pixel_row.append(palette[(byte & 128) // 128])
                    byte *= 2
        pixels.append(pixel_row[:width])
        return pixels

    def _get_pixels_from_udg_array(self, bit_depth, palette, udg_array, scale, mask, x, y, width, height):
        full_width = 8 * len(udg_array[0]) * scale
        full_height = 8 * len(udg_array) * scale
        width = min(width or full_width, full_width - x)
        height = min(height or full_height, full_height - y)

        t = [TRANSPARENT] * scale if TRANSPARENT in palette else None
        pixels = []
        y_count = 0
        for row in udg_array:
            if y_count >= y + height:
                break
            for j in range(8):
                if y <= y_count:
                    num_lines = min(y + height - y_count, scale)
                elif y - y_count <= scale:
                    num_lines = y - y_count
                else:
                    num_lines = 0
                y_count += scale
                if num_lines:
                    pixel_row = []
                    x_count = 0
                    for udg in row:
                        paper, ink = ATTR_INDEX[udg.attr & 127]
                        p_rgb = PALETTE[paper]
                        i_rgb = PALETTE[ink]
                        p = [p_rgb] * scale if p_rgb in palette else None
                        i = [i_rgb] * scale if i_rgb in palette else None
                        byte = udg.data[j]
                        if mask and udg.mask:
                            mask_byte = udg.mask[j]
                            byte &= mask_byte
                        else:
                            mask_byte = 0
                        for k in range(8):
                            if x <= x_count:
                                num_bits = scale
                            elif x - x_count <= scale:
                                num_bits = x - x_count
                            else:
                                num_bits = 0
                            x_count += scale
                            if num_bits:
                                if byte & 128:
                                    if i:
                                        pixel_row.extend(i)
                                    else:
                                        self.fail('Found ink colour %s not in expected palette: %s' % (i_rgb, palette))
                                elif mask_byte & 128:
                                    if t:
                                        pixel_row.extend(t)
                                    else:
                                        self.fail('Found transparent colour %s not in expected palette: %s' % (TRANSPARENT, palette))
                                else:
                                    if p:
                                        pixel_row.extend(p)
                                    else:
                                        self.fail('Found paper colour %s not in expected palette: %s' % (p_rgb, palette))
                            byte *= 2
                            mask_byte *= 2
                    pixel_row = pixel_row[:width]
                    for n in range(scale):
                        pixels.append(pixel_row)
                if y_count >= y + height:
                    break
        return pixels

    def _check_image_data(self, bit_depth, palette, image_data, exp_pixels, udg_array, scale, mask, x, y, width, height):
        pixels = self._get_pixels_from_image_data(bit_depth, palette, image_data, width)
        if exp_pixels is None:
            exp_pixels = self._get_pixels_from_udg_array(bit_depth, palette, udg_array, scale, mask, x, y, width, height)
        self.assertEqual(pixels, exp_pixels)

    def _check_idat(self, img_bytes, index, bit_depth, palette, exp_pixels, udg_array, scale, mask, x, y, width, height):
        i = index
        i, chunk_length = self._get_dword(img_bytes, i)
        idat_start = i
        i, chunk_type = self._get_chunk_type(img_bytes, i)
        self.assertEqual(chunk_type, IDAT)
        if PY2:
            image_data_z = ''.join(chr(b) for b in img_bytes[i:i + chunk_length])
            image_data = [ord(c) for c in zlib.decompress(image_data_z)]
        else:
            image_data_z = bytes(img_bytes[i:i + chunk_length])
            image_data = list(zlib.decompress(image_data_z))
        self._check_image_data(bit_depth, palette, image_data, exp_pixels, udg_array, scale, mask, x, y, width, height)
        i += chunk_length
        idat_end = i
        i, idat_crc = self._get_dword(img_bytes, i)
        self.assertEqual(idat_crc, self._get_crc(img_bytes[idat_start:idat_end]))
        return i

    def _get_image_data(self, image_writer, udg_array, scale=1, mask=False, x=0, y=0, width=None, height=None):
        img_stream = StreamIO()
        image_writer.write_image(udg_array, img_stream, scale, mask, x, y, width, height)
        if PY2:
            img_bytes = [ord(c) for c in img_stream.getvalue()]
        else:
            img_bytes = [b for b in img_stream.getvalue()]
        img_stream.close()
        return img_bytes

    def _test_image(self, udg_array, exp_bit_depth, exp_palette, exp_pixels=None, scale=1, mask=False, x=0, y=0, width=None, height=None, iw_args=None):
        image_writer = get_writer(**(iw_args or {}))
        img_bytes = self._get_image_data(image_writer, udg_array, scale, mask, x, y, width, height)

        # PNG signature
        i = self._check_signature(img_bytes)

        # IHDR
        exp_width = 8 * scale * len(udg_array[0]) - x if width is None else width
        exp_height = 8 * scale * len(udg_array) - y if height is None else height
        i, width, bit_depth = self._check_ihdr(img_bytes, i, exp_width, exp_height, exp_bit_depth)

        # PLTE
        i, palette = self._check_plte(img_bytes, i, exp_palette)

        # IDAT
        i = self._check_idat(img_bytes, i, bit_depth, palette, exp_pixels, udg_array, scale, mask, x, y, width, height)

        # IEND
        self.assertEqual(img_bytes[i:], IEND)

    def test_default_option_values(self):
        image_writer = get_writer()
        self.assertEqual(image_writer.options[PNG_COMPRESSION_LEVEL], 9)
        self.assertEqual(image_writer.options[PNG_MINIMISE_BIT_DEPTH], 1)
        self.assertEqual(image_writer.options[PNG_USE_FAST_METHOD], 1)

    def test_invalid_option_value(self):
        image_writer = get_writer(options={PNG_MINIMISE_BIT_DEPTH: 'NaN'})
        self.assertEqual(image_writer.options[PNG_MINIMISE_BIT_DEPTH], 1)

    def test_masked_fp(self):
        # Masked image, fixed palette (bit depth 4)
        iw_args = {'options': {PNG_MINIMISE_BIT_DEPTH: 0}}
        udg = Udg(56, (170,) * 8, (221,) * 8)
        udg_array = [[udg]]
        exp_palette = PALETTE
        self._test_image(udg_array, 4, exp_palette, mask=True, iw_args=iw_args)

    def test_masked_rp_2_colours_4_udgs_ink(self):
        # Masked image, reduced palette, two colours (trans + ink), >= 4 UDGs
        udg = Udg(49, (136,) * 8, (255,) * 8)
        udg_array = [[udg, udg]] * 2
        exp_palette = [TRANSPARENT, BLUE]
        self._test_image(udg_array, 1, exp_palette, mask=True)

    def test_masked_rp_2_colours_4_udgs_paper(self):
        # Masked image, reduced palette, two colours (trans + paper), >= 4 UDGs
        udg = Udg(8, (0,) * 8, (88,) * 8)
        udg_array = [[udg, udg]] * 2
        exp_palette = [TRANSPARENT, BLUE]
        self._test_image(udg_array, 1, exp_palette,  mask=True)

    def test_masked_rp_2_colours(self):
        # Masked image, reduced palette, two colours, < 4 UDGs
        udg = Udg(49, (136,) * 8, (255,) * 8)
        udg_array = [[udg]]
        exp_palette = [TRANSPARENT, BLUE]
        self._test_image(udg_array, 1, exp_palette, mask=True)

    def test_masked_rp_all_colours(self):
        # Masked image, reduced palette, 15 colours + trans
        udg1 = Udg(56, (15,) * 8)             # white, black
        udg2 = Udg(10, (15,) * 8)             # blue, red
        udg3 = Udg(28, (15,) * 8)             # magenta, green
        udg4 = Udg(46, (15,) * 8)             # cyan, yellow
        udg5 = Udg(120, (0,) * 8, (240,) * 8) # trans, bright white
        udg6 = Udg(74, (15,) * 8)             # bright blue, bright red
        udg7 = Udg(92, (15,) * 8)             # bright magenta, bright green
        udg8 = Udg(110, (15,) * 8)            # bright cyan, bright yellow
        udg_array = [[udg1, udg2, udg3, udg4, udg5, udg6, udg7, udg8]]
        exp_palette = [TRANSPARENT, BLACK, BLUE, RED, MAGENTA, GREEN, CYAN, YELLOW, WHITE, BRIGHT_BLUE, BRIGHT_RED, BRIGHT_MAGENTA, BRIGHT_GREEN, BRIGHT_CYAN, BRIGHT_YELLOW, BRIGHT_WHITE]
        self._test_image(udg_array, 4, exp_palette, mask=True)

    def test_masked_rp_bd2_generic_truncated(self):
        # Masked image, reduced palette, bit depth 2, generic method, truncated
        iw_args = {'options': {PNG_USE_FAST_METHOD: 0}}
        udg = Udg(88, (34,) * 8, (119,) * 8)
        udg_array = [[udg] * 3]
        exp_palette = [TRANSPARENT, BLACK, BRIGHT_MAGENTA]
        self._test_image(udg_array, 2, exp_palette, mask=True, width=17, iw_args=iw_args)

    def test_masked_rp_bd2(self):
        # Masked image, reduced palette, bit depth 2
        udg = Udg(88, (34,) * 8, (119,) * 8)
        udg_array = [[udg]]
        exp_palette = [TRANSPARENT, BLACK, BRIGHT_MAGENTA]
        self._test_image(udg_array, 2, exp_palette, mask=True)

    def test_masked_rp_bd4_all_bp(self):
        # Masked image, reduced palette, bit depth 4, all bit patterns
        udg1 = Udg(56, (0,) * 8)              # pppppppp
        udg2 = Udg(56, (255,) * 8)            # iiiiiiii
        udg3 = Udg(50, (27,) * 8)             # pp,pi,ip,ii
        udg4 = Udg(50, (0,) * 8, (228,) * 8)  # tt,tp,pt,pp
        udg5 = Udg(50, (27,) * 8, (255,) * 8) # tt,ti,it,ii
        udg6 = Udg(50, (6,) * 8, (246,) * 8)  # tt,tt,pi,ip
        udg_array = [[udg1, udg2, udg3, udg4, udg5, udg6]]
        exp_palette = [TRANSPARENT, WHITE, BLACK, RED, YELLOW]
        self._test_image(udg_array, 4, exp_palette, mask=True)

    def test_masked_rp_bd4_generic(self):
        # Masked image, reduced palette, bit depth 4, generic method
        iw_args = {'options': {PNG_USE_FAST_METHOD: 0}}
        udg1 = Udg(24, (34,) * 8, (119,) * 8)
        udg2 = Udg(57, (170,) * 8)
        udg_array = [[udg1, udg2]]
        exp_palette = [TRANSPARENT, WHITE, BLACK, BLUE, MAGENTA]
        self._test_image(udg_array, 4, exp_palette, mask=True, iw_args=iw_args)

    def test_masked_rp_single_colour_scale1(self):
        # Masked image, reduced palette, single colour, scale 1
        udg = Udg(56, (0,) * 8, (255,) * 8)
        udg_array = [[udg]]
        exp_palette = [TRANSPARENT]
        self._test_image(udg_array, 1, exp_palette, mask=True)

    def test_masked_rp_single_colour_scale2(self):
        # Masked image, reduced palette, single colour, scale 2
        udg = Udg(56, (0,) * 8, (255,) * 8)
        udg_array = [[udg]]
        exp_palette = [TRANSPARENT]
        self._test_image(udg_array, 1, exp_palette, scale=2, mask=True)

    def test_no_build(self):
        image_writer = get_writer(build_images=False)
        img_data = self._get_image_data(image_writer, [[Udg(0, (0,) * 8)]])
        self.assertEqual(len(img_data), 0)

    def test_unmasked_fp_scale2(self):
        # Unmasked image, fixed palette (bit depth 4), scale 2
        iw_args = {'options': {PNG_MINIMISE_BIT_DEPTH: 0}}
        udg = Udg(56, (128, 64, 32, 16, 8, 4, 2, 1))
        udg_array = [[udg]]
        exp_palette = PALETTE
        self._test_image(udg_array, 4, exp_palette, scale=2, iw_args=iw_args)

    def test_unmasked_fp(self):
        # Unmasked image, fixed palette (bit depth 4)
        iw_args = {'options': {PNG_MINIMISE_BIT_DEPTH: 0}}
        udg = Udg(56, (128, 64, 32, 16, 8, 4, 2, 1))
        udg_array = [[udg]]
        exp_palette = PALETTE
        self._test_image(udg_array, 4, exp_palette, iw_args=iw_args)

    def test_unmasked_fp_truncated(self):
        # Unmasked image, fixed palette (bit depth 4), truncated width
        iw_args = {'options': {PNG_MINIMISE_BIT_DEPTH: 0}}
        udg = Udg(56, (128, 64, 32, 16, 8, 4, 2, 1))
        udg_array = [[udg, udg]]
        exp_palette = PALETTE
        self._test_image(udg_array, 4, exp_palette, width=13, iw_args=iw_args)

    def test_unmasked_rp_2_colours_4_udgs(self):
        # Unmasked image, reduced palette, two colours, 4 UDGs
        udg = Udg(49, (136,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [BLUE, YELLOW]
        self._test_image(udg_array, 1, exp_palette)

    def test_unmasked_rp_2_custom_colours(self):
        # Unmasked image, reduced palette, two custom colours
        indigo = [75, 0, 130]
        salmon = [250, 128, 114]
        colours = {'BLUE': indigo, 'RED': salmon}
        iw_args = {'palette': colours}
        udg = Udg(10, (195,) * 8) # 11000011
        udg_array = [[udg]]
        exp_palette = [salmon, indigo]
        exp_pixels = [[salmon, salmon, indigo, indigo, indigo, indigo, salmon, salmon]] * 8
        self._test_image(udg_array, 1, exp_palette, exp_pixels=exp_pixels, iw_args=iw_args)

    def test_unmasked_rp_all_colours(self):
        # Unmasked image, reduced palette, 15 colours
        udg1 = Udg(56, (15,) * 8)  # white, black
        udg2 = Udg(10, (15,) * 8)  # blue, red
        udg3 = Udg(28, (15,) * 8)  # magenta, green
        udg4 = Udg(46, (15,) * 8)  # cyan, yellow
        udg5 = Udg(120, (0,) * 8)  # bright white
        udg6 = Udg(74, (15,) * 8)  # bright blue, bright red
        udg7 = Udg(92, (15,) * 8)  # bright magenta, bright green
        udg8 = Udg(110, (15,) * 8) # bright cyan, bright yellow
        udg_array = [[udg1, udg2, udg3, udg4, udg5, udg6, udg7, udg8]]
        exp_palette = [BLACK, BLUE, RED, MAGENTA, GREEN, CYAN, YELLOW, WHITE, BRIGHT_BLUE, BRIGHT_RED, BRIGHT_MAGENTA, BRIGHT_GREEN, BRIGHT_CYAN, BRIGHT_YELLOW, BRIGHT_WHITE]
        self._test_image(udg_array, 4, exp_palette)

    def test_unmasked_rp_bd1_cropped(self):
        # Unmasked image, reduced palette, bit depth 1, cropped all sides
        udg = Udg(30, (170,) * 8)
        udg_array = [[udg] * 2] * 2
        exp_palette = [YELLOW, MAGENTA]
        self._test_image(udg_array, 1, exp_palette, x=1, y=2, width=9, height=11)

    def test_unmasked_rp_bd1_cropped_bottom(self):
        # Unmasked image, reduced palette, bit depth 1, cropped bottom
        udg = Udg(28, (148,) * 8)
        udg_array = [[udg] * 2] * 2
        exp_palette = [MAGENTA, GREEN]
        self._test_image(udg_array, 1, exp_palette, height=12)

    def test_unmasked_rp_bd1_cropped_left(self):
        # Unmasked image, reduced palette, bit depth 1, cropped left
        udg = Udg(17, (148,) * 8)
        udg_array = [[udg] * 2] * 2
        exp_palette = [BLUE, RED]
        self._test_image(udg_array, 1, exp_palette, x=5)

    def test_unmasked_rp_bd1_cropped_right(self):
        # Unmasked image, reduced palette, bit depth 1, cropped right
        udg = Udg(5, (148,) * 8)
        udg_array = [[udg] * 2] * 2
        exp_palette = [BLACK, CYAN]
        self._test_image(udg_array, 1, exp_palette, width=11)

    def test_unmasked_rp_bd1_cropped_top(self):
        # Unmasked image, reduced palette, bit depth 1, cropped top
        udg = Udg(55, (148,) * 8)
        udg_array = [[udg] * 2] * 2
        exp_palette = [WHITE, YELLOW]
        self._test_image(udg_array, 1, exp_palette, y=7)

    def test_unmasked_rp_bd1_scale2(self):
        # Unmasked image, reduced palette, bit depth 1, scale 2, >3 UDGs
        udg = Udg(49, (164,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [BLUE, YELLOW]
        self._test_image(udg_array, 1, exp_palette, scale=2)

    def test_unmasked_rp_bd1_scale3(self):
        # Unmasked image, reduced palette, bit depth 1, scale 3, >3 UDGs
        udg = Udg(42, (91,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [RED, CYAN]
        self._test_image(udg_array, 1, exp_palette, scale=3)

    def test_unmasked_rp_bd1_scale4(self):
        # Unmasked image, reduced palette, bit depth 1, scale 4, >3 UDGs
        udg = Udg(24, (137,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [BLACK, MAGENTA]
        self._test_image(udg_array, 1, exp_palette, scale=4)

    def test_unmasked_rp_bd1_scale5(self):
        # Unmasked image, reduced palette, bit depth 1, scale 5, >3 UDGs
        udg = Udg(16, (27,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [BLACK, RED]
        self._test_image(udg_array, 1, exp_palette, scale=5)

    def test_unmasked_rp_bd1_scale6(self):
        # Unmasked image, reduced palette, bit depth 1, scale 6, >3 UDGs
        udg = Udg(46, (138,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [CYAN, YELLOW]
        self._test_image(udg_array, 1, exp_palette, scale=6)

    def test_unmasked_rp_bd1_scale7(self):
        # Unmasked image, reduced palette, bit depth 1, scale 7, >3 UDGs
        udg = Udg(59, (75,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [WHITE, MAGENTA]
        self._test_image(udg_array, 1, exp_palette, scale=7)

    def test_unmasked_rp_bd1_scale8(self):
        # Unmasked image, reduced palette, bit depth 1, scale 8, >3 UDGs
        udg = Udg(1, (41,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [BLACK, BLUE]
        self._test_image(udg_array, 1, exp_palette, scale=8)

    def test_unmasked_rp_bd1_scale9(self):
        # Unmasked image, reduced palette, bit depth 1, scale 9, >3 UDGs
        udg = Udg(61, (101,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [WHITE, CYAN]
        self._test_image(udg_array, 1, exp_palette, scale=9)

    def test_unmasked_rp_bd1_scale10(self):
        # Unmasked image, reduced palette, bit depth 1, scale 10, >3 UDGs
        udg = Udg(50, (249,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [RED, YELLOW]
        self._test_image(udg_array, 1, exp_palette, scale=10)

    def test_unmasked_rp_bd1_scale11(self):
        # Unmasked image, reduced palette, bit depth 1, scale 11, >3 UDGs
        udg = Udg(44, (183,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [GREEN, CYAN]
        self._test_image(udg_array, 1, exp_palette, scale=11)

    def test_unmasked_rp_bd1_scale12(self):
        # Unmasked image, reduced palette, bit depth 1, scale 12, >3 UDGs
        udg = Udg(30, (221,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [MAGENTA, YELLOW]
        self._test_image(udg_array, 1, exp_palette, scale=12)

    def test_unmasked_rp_bd1_scale13(self):
        # Unmasked image, reduced palette, bit depth 1, scale 13, >3 UDGs
        udg = Udg(5, (148,) * 8)
        udg_array = [[udg] * 4]
        exp_palette = [BLACK, CYAN]
        self._test_image(udg_array, 1, exp_palette, scale=13)

    def test_unmasked_rp_bd1_generic(self):
        # Unmasked image, reduced palette, bit depth 1, generic method
        iw_args = {'options': {PNG_USE_FAST_METHOD: 0}}
        udg = Udg(5, (148,) * 8)
        udg_array = [[udg]]
        exp_palette = [BLACK, CYAN]
        self._test_image(udg_array, 1, exp_palette, iw_args=iw_args)

    def test_unmasked_rp_bd2_cropped(self):
        # Unmasked image, reduced palette, bit depth 2, cropped all sides
        udg1 = Udg(30, (170,) * 8)
        udg2 = Udg(28, (81,) * 8)
        udg_array = [[udg1, udg2]] * 2
        exp_palette = [YELLOW, MAGENTA, GREEN]
        self._test_image(udg_array, 2, exp_palette, x=1, y=2, width=9, height=11)

    def test_unmasked_rp_bd2_cropped_bottom(self):
        # Unmasked image, reduced palette, bit depth 2, cropped bottom
        udg1 = Udg(28, (148,) * 8)
        udg2 = Udg(17, (148,) * 8)
        udg_array = [[udg1, udg2]] * 2
        exp_palette = [MAGENTA, GREEN, BLUE, RED]
        self._test_image(udg_array, 2, exp_palette, height=12)

    def test_unmasked_rp_bd2_cropped_left(self):
        # Unmasked image, reduced palette, bit depth 2, cropped left
        udg1 = Udg(17, (148,) * 8)
        udg2 = Udg(5, (148,) * 8)
        udg_array = [[udg1, udg2]] * 2
        exp_palette = [BLUE, RED, BLACK, CYAN]
        self._test_image(udg_array, 2, exp_palette, x=5)

    def test_unmasked_rp_bd2_cropped_right(self):
        # Unmasked image, reduced palette, bit depth 2, cropped right
        udg1 = Udg(5, (148,) * 8)
        udg2 = Udg(55, (81,) * 8)
        udg_array = [[udg1, udg2]] * 2
        exp_palette = [BLACK, CYAN, WHITE, YELLOW]
        self._test_image(udg_array, 2, exp_palette, width=11)

    def test_unmasked_rp_bd2_cropped_top(self):
        # Unmasked image, reduced palette, bit depth 2, cropped top
        udg1 = Udg(55, (148,) * 8)
        udg2 = Udg(4, (148,) * 8)
        udg_array = [[udg1, udg2]] * 2
        exp_palette = [WHITE, YELLOW, BLACK, GREEN]
        self._test_image(udg_array, 2, exp_palette, y=7)

    def test_unmasked_rp_bd2_scale2(self):
        # Unmasked image, reduced palette, bit depth 2, scale 2
        udg1 = Udg(56, (93,) * 8)
        udg2 = Udg(49, (162,) * 8)
        udg_array = [[udg1, udg2]]
        exp_palette = [WHITE, BLACK, BLUE, YELLOW]
        self._test_image(udg_array, 2, exp_palette, scale=2)

    def test_unmasked_rp_bd2_scale3(self):
        # Unmasked image, reduced palette, bit depth 2, scale 3
        udg1 = Udg(4, (195,) * 8)
        udg2 = Udg(33, (105,) * 8)
        udg_array = [[udg1, udg2]]
        exp_palette = [BLACK, BLUE, GREEN]
        self._test_image(udg_array, 2, exp_palette, scale=3)

    def test_unmasked_rp_bd2_scale4(self):
        # Unmasked image, reduced palette, bit depth 2, scale 4
        udg1 = Udg(32, (33,) * 8)
        udg2 = Udg(10, (135,) * 8)
        udg_array = [[udg1, udg2]]
        exp_palette = [BLACK, BLUE, RED, GREEN]
        self._test_image(udg_array, 2, exp_palette, scale=4)

    def test_unmasked_rp_bd2_scale5(self):
        # Unmasked image, reduced palette, bit depth 2, scale 5
        udg1 = Udg(55, (249,) * 8)
        udg2 = Udg(44, (83,) * 8)
        udg_array = [[udg1, udg2]]
        exp_palette = [WHITE, GREEN, CYAN, YELLOW]
        self._test_image(udg_array, 2, exp_palette, scale=5)

    def test_unmasked_rp_bd2_truncated(self):
        # Unmasked image, reduced palette, bit depth 2, truncated
        udg1 = Udg(5, (148,) * 8)
        udg2 = Udg(19, (107,) * 8)
        udg_array = [[udg1, udg2] * 3]
        exp_palette = [BLACK, RED, MAGENTA, CYAN]
        self._test_image(udg_array, 2, exp_palette, width=36)

    def test_unmasked_rp_bd2(self):
        # Unmasked image, reduced palette, bit depth 2
        udg1 = Udg(49, (136,) * 8)
        udg2 = Udg(42, (68,) * 8)
        udg_array = [[udg1, udg2]]
        exp_palette = [BLUE, RED, CYAN, YELLOW]
        self._test_image(udg_array, 2, exp_palette)

    def test_unmasked_rp_bd4_cropped(self):
        # Unmasked image, reduced palette, bit depth 4, cropped all sides
        udg1 = Udg(30, (170,) * 8)
        udg2 = Udg(28, (81,) * 8)
        udg3 = Udg(5, (129,) * 8)
        udg_array = [[udg1, udg2], [udg3, udg1]]
        exp_palette = [YELLOW, MAGENTA, GREEN, CYAN, BLACK]
        self._test_image(udg_array, 4, exp_palette, x=2, y=1, width=11, height=9)

    def test_unmasked_rp_bd4_cropped_bottom(self):
        # Unmasked image, reduced palette, bit depth 4, cropped bottom
        udg1 = Udg(28, (148,) * 8)
        udg2 = Udg(17, (148,) * 8)
        udg3 = Udg(5, (195,) * 8)
        udg_array = [[udg1, udg2], [udg3, udg1]]
        exp_palette = [MAGENTA, GREEN, BLUE, RED, CYAN, BLACK]
        self._test_image(udg_array, 4, exp_palette, height=7)

    def test_unmasked_rp_bd4_cropped_left(self):
        # Unmasked image, reduced palette, bit depth 4, cropped left
        udg1 = Udg(17, (148,) * 8)
        udg2 = Udg(5, (148,) * 8)
        udg3 = Udg(22, (1,) * 8)
        udg_array = [[udg1, udg2], [udg3, udg1]]
        exp_palette = [BLUE, RED, BLACK, CYAN, YELLOW]
        self._test_image(udg_array, 4, exp_palette, x=3)

    def test_unmasked_rp_bd4_cropped_right(self):
        # Unmasked image, reduced palette, bit depth 4, cropped right
        udg1 = Udg(5, (148,) * 8)
        udg2 = Udg(55, (81,) * 8)
        udg3 = Udg(76, (67,) * 8)
        udg_array = [[udg1, udg2], [udg3, udg1]]
        exp_palette = [BLACK, CYAN, WHITE, YELLOW, BRIGHT_BLUE, BRIGHT_GREEN]
        self._test_image(udg_array, 4, exp_palette, width=11)

    def test_unmasked_rp_bd4_cropped_top(self):
        # Unmasked image, reduced palette, bit depth 4, cropped top
        udg1 = Udg(55, (148,) * 8)
        udg2 = Udg(4, (148,) * 8)
        udg3 = Udg(42, (63,) * 8)
        udg_array = [[udg1, udg2], [udg3, udg1]]
        exp_palette = [WHITE, YELLOW, BLACK, GREEN, CYAN, RED]
        self._test_image(udg_array, 4, exp_palette, y=7)

    def test_unmasked_rp_bd4(self):
        # Unmasked image, reduced palette, bit depth 4
        udg1 = Udg(49, (136,) * 8)
        udg2 = Udg(42, (68,) * 8)
        udg3 = Udg(35, (34,) * 8)
        udg_array = [[udg1, udg2, udg3]]
        exp_palette = [BLUE, RED, MAGENTA, GREEN, CYAN, YELLOW]
        self._test_image(udg_array, 4, exp_palette)

    def test_unmasked_rp(self):
        # Unmasked image, reduced palette
        udg = Udg(56, (128, 64, 32, 16, 8, 4, 2, 1))
        udg_array = [[udg]]
        exp_palette = [WHITE, BLACK]
        self._test_image(udg_array, 1, exp_palette)

    def test_unmasked_rp_single_colour(self):
        # Unmasked image, reduced palette, single colour
        udg = Udg(56, (255,) * 8)
        udg_array = [[udg]]
        exp_palette = [BLACK]
        self._test_image(udg_array, 1, exp_palette)

    def test_unmasked_rp_single_colour_scale2(self):
        # Unmasked image, reduced palette, single colour, scale 2
        udg = Udg(57, (255,) * 8)
        udg_array = [[udg]]
        exp_palette = [BLUE]
        self._test_image(udg_array, 1, exp_palette, scale=2)

if __name__ == '__main__':
    unittest.main()
